from typing import List, Text

import regex
from ruth.nlu.tokenizer.tokenizer import Token, Tokenizer


class WhiteSpaceTokenizer(Tokenizer):
    def __init__(self, element_config):
        super().__init__(element_config)

    def process(self):
        pass

    def tokenize(self, text: Text) -> List[Token]:
        words = regex.sub(
            # there is a space or an end of a string after it
            r"(www\.[^\s]+)|(https?://[^\s]+)|(pic\.twitter\.com/[^\s]+)"
            r"[^\w#@&]+(?=\s|$)|"
            # there is a space or beginning of a string before it
            # not followed by a number
            r"(\s|^)[^\w#@&]+(?=[^0-9\s])|"
            # not in between numbers and not . or @ or & or - or #
            # e.g. 10'000.00 or blabla@gmail.com
            # and not url characters
            r"(?<=[^0-9\s])[^\w._~:/?#\[\]()@!$&*+,;=-]+(?=[^0-9\s])",
            " ",
            text,
        ).split()
        tokens = self._convert_words_to_tokens(words, text)
        return tokens



import logging
from pathlib import Path
from typing import Any, Dict, List, Text, Tuple, Union

import sklearn
from numpy import argsort, fliplr, ndarray, reshape
from rich.console import Console
from ruth.constants import INTENT, INTENT_RANKING
from ruth.nlu.classifiers import LABEL_RANKING_LIMIT
from ruth.nlu.classifiers.ruth_classifier import IntentClassifier
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_data import RuthData
from ruth.shared.utils import json_pickle, json_unpickle
from scipy import sparse
from sklearn.model_selection import GridSearchCV
from sklearn.preprocessing import LabelEncoder

logger = logging.getLogger(__name__)

console = Console()


class SVMClassifier(IntentClassifier):
    defaults = {
        "C": [1, 2, 5, 10, 20, 100],
        "kernel": ["linear", "rbf"],
        "gamma": ["auto", 0.1],
        "decision_function_shape": ["ovr"],
        "max_cross_validation_folds": 5,
        "scoring": "f1_weighted",
        "max_length": 30000,
    }

    def __init__(
        self,
        element_config: Dict[Text, Any],
        le: LabelEncoder = None,
        clf: GridSearchCV = None,
    ):
        self.clf = clf
        super().__init__(element_config, le)

    @staticmethod
    def get_features(message: RuthData) -> Union[sparse.spmatrix, ndarray]:
        feature = message.get_features()
        if feature is not None:
            return feature.feature[0]
        raise ValueError("There is no sentence. Not able to train SVMClassifier")

    @property
    def param_grids(self):
        return {
            "C": self.element_config["C"],
            "kernel": self.element_config["kernel"],
            "gamma": self.element_config["gamma"],
            "decision_function_shape": self.element_config["decision_function_shape"],
        }

    def _create_gridsearch(self, X, y) -> "sklearn.model_selection.GridSearchCV":
        from sklearn.svm import SVC

        clf = SVC(probability=True)
        param_grids = self.param_grids
        return GridSearchCV(
            clf,
            param_grids,
            scoring=self.element_config["scoring"],
            # cv=self.element_config["max_cross_validation_folds"],
        )

    def train(self, training_data: TrainData):
        intents: List[Text] = [
            message.get(INTENT) for message in training_data.intent_examples
        ]
        if len(set(intents)) < 2:
            logger.warning(
                "There are no enough intent. "
                "At least two unique intent are needed to train the model"
            )
            return
        X = [self.get_features(message) for message in training_data.intent_examples]
        if self.check_dense(X[0]):

            max_length = self.get_max_length(X)
            self.element_config["max_length"] = max_length
            X = [self.ravel_vector(x) for x in X]
            # X = [self.pad_vector(x, max_length) for x in X]
        else:
            X = [message.toarray() for message in X]

        y = self.encode_the_str_to_int(intents)

        X = reshape(X, (len(X), -1))
        self.clf = self._create_gridsearch(X, y)
        self.clf.fit(X, y)
        console.print(f"The Best parameter we got are {self.clf.best_params_}")
        console.print(f"score: {self.clf.best_score_}")

    def persist(self, file_name: Text, model_dir: Path):
        classifier_file_name = file_name + "_classifier.pkl"
        encoder_file_name = file_name + "_encoder.pkl"

        classifier_path = model_dir / classifier_file_name
        encoder_path = model_dir / encoder_file_name

        if self.clf and self.le:
            json_pickle(classifier_path, self.clf.best_estimator_)
            json_pickle(encoder_path, self.le)

        return {"classifier": classifier_file_name, "encoder": encoder_file_name}

    def _predict(self, x: ndarray) -> Tuple[ndarray, ndarray]:
        predictions = self.predict_probabilities(x)
        sorted_index = fliplr(argsort(predictions, axis=1))
        return sorted_index, predictions[:, sorted_index]

    def predict_probabilities(self, x: ndarray) -> ndarray:
        return self.clf.predict_proba(x.reshape(1, -1))

    @classmethod
    def load(cls, meta: Dict[Text, Any], model_dir: Path):
        classifier_file_name = model_dir / meta["classifier"]
        encoder_file_name = model_dir / meta["encoder"]

        clf = json_unpickle(Path(classifier_file_name))
        le = json_unpickle(Path(encoder_file_name))

        return cls(meta, clf=clf, le=le)

    def parse(self, message: RuthData):
        x = self.get_features(message)
        if self.check_dense(x):
            x = self.ravel_vector(x)
            x = self.pad_vector(x, self.element_config["max_length"])
        else:
            x = x.toarray()
        index, probabilities = self._predict(x)

        intents = self._change_int_to_text(index.flatten())
        probabilities = probabilities.flatten()

        if intents.size > 0 and probabilities.size > 0:
            ranking = list(zip(list(intents), list(probabilities)))[
                :LABEL_RANKING_LIMIT
            ]
            intent = {"name": intents[0], "accuracy": probabilities[0]}
            intent_rankings = [
                {"name": name, "accuracy": probability} for name, probability in ranking
            ]
        else:
            intent = {"name": None, "accuracy": 0.0}
            intent_rankings = []
        message.set(INTENT, intent)
        message.set(INTENT_RANKING, intent_rankings)



from pathlib import Path

from ruth.nlu.constants import DEFAULT_MODEL_NAME
from ruth.nlu.model import Trainer
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_config import RuthConfig


def train_pipeline(config: RuthConfig, training_data: TrainData) -> Path:
    trainer = Trainer(config)
    trainer.train(training_data)

    model_dir = Path(DEFAULT_MODEL_NAME)
    model_dir.mkdir(exist_ok=True)
    model_absolute_path = trainer.persist(model_dir)
    return model_absolute_path



from typing import Any, Dict, Text, Union

import numpy as np
from numpy import ndarray
from ruth.constants import TEXT
from scipy import sparse


class Feature:
    def __init__(self, feature: Union[sparse.spmatrix, ndarray], origin: Text):
        self.feature = feature
        self.origin = origin

    def is_sparse(self):
        return isinstance(self.feature, sparse.spmatrix)

    def is_dense(self):
        return not self.is_sparse()

    def _combine_sparse_features(
        self, additional_features: "Feature", message: Dict[Text, Any]
    ) -> None:
        from scipy.sparse import hstack

        if self.feature.shape[0] != additional_features.feature.shape[0]:
            raise ValueError(
                f"Cannot combine sparse features as sequence dimensions do not "
                f"match: {self.feature.shape[0]} != "
                f"{additional_features.feature.shape[0]}."
            )

        self.features = hstack([self.feature, additional_features.feature])

    def _combine_dense_features(
        self, additional_features: Any, message: Dict[Text, Any]
    ) -> Any:
        if len(self.feature.shape[0]) != len(additional_features.feature.shape[0]):
            raise ValueError(
                f"Cannot concatenate dense features as sequence dimension does not "
                f"match: {self.feature.shape[0]} != "
                f"{len(additional_features)}. Message: {message.get(TEXT, 'Text not available')}"
            )
        else:
            return np.concatenate((self.feature, additional_features), axis=-1)

    def combine_with_features(
        self, additional_features: "Feature", message: Dict[Text, Any]
    ) -> None:
        if additional_features is None:
            return

        if self.is_dense() and additional_features.is_dense():
            self._combine_dense_features(additional_features, message)
        elif self.is_sparse() and additional_features.is_sparse():
            self._combine_sparse_features(additional_features, message)
        else:
            raise ValueError("Cannot combine sparse and dense features.")



import logging
from pathlib import Path
from typing import Any, Dict, Text, Tuple

import sklearn
from numpy import argsort, fliplr, ndarray, reshape
from ruth.constants import INTENT, INTENT_RANKING
from ruth.nlu.classifiers import LABEL_RANKING_LIMIT
from ruth.nlu.classifiers.ruth_classifier import IntentClassifier
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_data import RuthData
from ruth.shared.utils import json_pickle, json_unpickle
from scipy import sparse
from sklearn.preprocessing import LabelEncoder

logger = logging.getLogger(__name__)


class NaiveBayesClassifier(IntentClassifier):
    defaults = {"priors": None, "var_smoothing": 1e-9}

    def __init__(
        self,
        element_config: Dict[Text, Any],
        le: LabelEncoder = None,
        model: "sklearn.naive_bayes.GaussianNB" = None,
    ):
        super(NaiveBayesClassifier, self).__init__(element_config=element_config, le=le)

        self.model = model

    def _create_classifier(self) -> "sklearn.naive_bayes.GaussianNB":
        from sklearn.naive_bayes import GaussianNB

        priors = self.element_config["priors"]
        var_smoothing = self.element_config["var_smoothing"]

        return GaussianNB(priors=priors, var_smoothing=var_smoothing)

    def train(self, training_data: TrainData):
        intents = [message.get(INTENT) for message in training_data.intent_examples]
        if len(set(intents)) < 2:
            logger.warning(
                "There are no enough intent. "
                "At least two unique intent are needed to train the model"
            )
            return

        X = [
            self.get_features(message).toarray()
            for message in training_data.intent_examples
        ]
        y = self.encode_the_str_to_int(intents)

        X = reshape(X, (len(X), -1))
        self.model = self._create_classifier()
        self.model.fit(X, y)

    def _predict(self, x: ndarray) -> Tuple[ndarray, ndarray]:
        predictions = self.predict_probabilities(x)
        sorted_index = fliplr(argsort(predictions, axis=1))
        return sorted_index, predictions[:, sorted_index]

    def predict_probabilities(self, x: ndarray) -> ndarray:
        return self.model.predict_proba(x.reshape(1, -1))

    def _change_int_to_text(self, prediction: ndarray) -> ndarray:

        return self.le.inverse_transform(prediction)

    @staticmethod
    def get_features(message: RuthData) -> sparse.spmatrix:
        feature = message.get_features()
        if feature is not None:
            return feature.feature[0]
        raise ValueError("There is no sentence. Not able to train NaiveBayesClassifier")

    def parse(self, message: RuthData):
        x = self.get_features(message).toarray()
        index, probabilities = self._predict(x)

        intents = self._change_int_to_text(index.flatten())
        probabilities = probabilities.flatten()

        if intents.size > 0 and probabilities.size > 0:
            ranking = list(zip(list(intents), list(probabilities)))[
                :LABEL_RANKING_LIMIT
            ]
            intent = {"name": intents[0], "accuracy": probabilities[0]}
            intent_rankings = [
                {"name": name, "accuracy": probability} for name, probability in ranking
            ]
        else:
            intent = {"name": None, "accuracy": 0.0}
            intent_rankings = []
        message.set(INTENT, intent)
        message.set(INTENT_RANKING, intent_rankings)

    def persist(self, file_name: Text, model_dir: Path):
        classifier_file_name = file_name + "_classifier.pkl"
        encoder_file_name = file_name + "_encoder.pkl"

        classifier_path = model_dir / classifier_file_name
        encoder_path = model_dir / encoder_file_name

        if self.model and self.le:
            json_pickle(classifier_path, self.model)
            json_pickle(encoder_path, self.le)

        return {"classifier": classifier_file_name, "encoder": encoder_file_name}

    @classmethod
    def load(cls, meta: Dict[Text, Any], model_dir: Path):
        classifier_file_name = model_dir / meta["classifier"]
        encoder_file_name = model_dir / meta["encoder"]

        model = json_unpickle(Path(classifier_file_name))
        le = json_unpickle(Path(encoder_file_name))

        return cls(meta, model=model, le=le)



import logging
from pathlib import Path
from typing import Any, Dict, List, Text

from ruth.shared.constants import ELEMENT_INDEX, KEY_NAME
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_data import RuthData
from ruth.shared.nlu.training_data.utils import override_defaults

logger = logging.getLogger(__name__)


class ElementMetaClass(type):
    @property
    def name(cls) -> Text:
        return cls.__name__


class Element(metaclass=ElementMetaClass):
    defaults = {}

    def __init__(self, element_config: Dict[Text, Any]):
        element_config = element_config or {}

        element_config[KEY_NAME] = self.name
        self.element_config = override_defaults(self.defaults, element_config)

    @property
    def name(self):
        return type(self).name

    def train(self, training_data: TrainData):
        pass

    def parse(self, message: RuthData):
        pass

    def create_unique_name(self) -> Text:
        idx = self.element_config.get(ELEMENT_INDEX)
        return self.name if idx is None else f"element_{idx}_{self.name}"

    @classmethod
    def build(cls, element_config: Dict[Text, Any]):
        return cls(element_config)

    def persist(self, file_name: Text, model_dir: Text):
        pass

    @classmethod
    def load(cls, meta: Dict[Text, Any], model_dir: Path, **kwargs: Any):
        return cls(meta)

    @staticmethod
    def required_element() -> List[object]:
        return []



import logging
from collections import Counter
from pathlib import Path
from typing import Any, Dict, List, Text, Tuple

import torch
from numpy import argsort, fliplr
from rich.console import Console
from ruth.constants import INTENT, INTENT_RANKING
from ruth.nlu.classifiers import LABEL_RANKING_LIMIT
from ruth.nlu.classifiers.constants import BATCH_SIZE, EPOCHS, MODEL_NAME
from ruth.nlu.classifiers.ruth_classifier import IntentClassifier
from ruth.nlu.tokenizer.hf_tokenizer import HFTokenizer
from ruth.shared.constants import ATTENTION_MASKS, INPUT_IDS
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_data import RuthData
from ruth.shared.utils import json_pickle, json_unpickle
from sklearn.preprocessing import LabelEncoder
from torch import nn
from torch.nn import Module
from torch.optim import AdamW
from torch.utils.data import DataLoader, Dataset
from tqdm import tqdm
from transformers import AutoModelForSequenceClassification
from transformers import logging as transformer_logging

torch.cuda.empty_cache()
logger = logging.getLogger(__name__)
transformer_logging.set_verbosity_error()

console = Console()


class HFClassifier(IntentClassifier):
    defaults = {EPOCHS: 100, MODEL_NAME: "bert-base-uncased", BATCH_SIZE: 1}

    def __init__(
        self,
        element_config: Dict[Text, Any],
        le: LabelEncoder = None,
        model: Module = None,
    ):
        self.model = model
        super().__init__(element_config, le)

    def required_element(self):
        return [HFTokenizer]

    def _build_model(self, label_count):
        return AutoModelForSequenceClassification.from_pretrained(
            self.element_config[MODEL_NAME], num_labels=label_count
        )

    @staticmethod
    def get_input_ids(message: RuthData) -> Dict[Text, List[int]]:
        input_ids = message.get(INPUT_IDS)
        if input_ids is not None:
            return input_ids
        raise ValueError("There is no sentence. Not able to train HFClassifier")

    @staticmethod
    def get_attention_masks(message: RuthData) -> Dict[Text, List[int]]:
        attention_masks = message.get(ATTENTION_MASKS)
        if attention_masks is not None:
            return attention_masks
        raise ValueError("There is no sentence. Not able to train HFClassifier")

    @staticmethod
    def get_optimizer(model):
        return AdamW(model.parameters(), lr=5e-5)

    @staticmethod
    def get_device():
        return (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )

    @property
    def get_params(self):
        return {
            EPOCHS: self.element_config[EPOCHS],
            BATCH_SIZE: self.element_config[BATCH_SIZE],
        }

    def train(self, training_data: TrainData):
        intents: List[Text] = [
            message.get(INTENT) for message in training_data.intent_examples
        ]
        if len(set(intents)) < 2:
            logger.warning(
                "There are no enough intent. "
                "At least two unique intent are needed to train the model"
            )
            return

        X = {
            "input_ids": [
                self.get_input_ids(message) for message in training_data.intent_examples
            ],
            "attention_masks": [
                self.get_attention_masks(message)
                for message in training_data.intent_examples
            ],
        }

        y = self.encode_the_str_to_int(intents)
        label_count = len(Counter(y).keys())

        params = self.get_params

        loaded_data = HFDatasetLoader(X, y)
        batched_data = DataLoader(
            loaded_data, batch_size=params[BATCH_SIZE], shuffle=True
        )

        self.model = self._build_model(label_count)

        optimizer = self.get_optimizer(self.model)
        device = self.get_device()
        logger.info("device: " + str(device) + " is used")
        self.model.to(device)

        self.model.train()
        for epoch in range(params[EPOCHS]):
            for batch in tqdm(batched_data, desc="epoch " + str(epoch)):
                optimizer.zero_grad()
                input_ids = batch["input_ids"].to(device)
                attention_masks = batch["attention_masks"].to(device)
                labels = batch["labels"].to(device)
                outputs = self.model(
                    input_ids, attention_mask=attention_masks, labels=labels
                )
                loss = outputs[0]
                loss.backward()
                optimizer.step()

    def persist(self, file_name: Text, model_dir: Path):
        classifier_file_name = file_name + "_classifier"
        encoder_file_name = file_name + "_encoder.pkl"

        classifier_path = str(model_dir) + "/" + classifier_file_name
        encoder_path = str(model_dir) + "/" + encoder_file_name

        if self.model and self.le:
            model_to_save = (
                self.model.module if hasattr(self.model, "module") else self.model
            )
            model_to_save.save_pretrained(classifier_path)
            json_pickle(encoder_path, self.le)

        return {"classifier": classifier_file_name, "encoder": encoder_file_name}

    @classmethod
    def load(cls, meta: Dict[Text, Any], model_dir: Path):
        classifier_file_name = model_dir / meta["classifier"]
        encoder_file_name = model_dir / meta["encoder"]

        classifier = AutoModelForSequenceClassification.from_pretrained(
            classifier_file_name
        )
        le = json_unpickle(Path(encoder_file_name))

        return cls(meta, model=classifier, le=le)

    def _predict(self, input_ids, attention_masks) -> Tuple[List[int], List[float]]:
        predictions = self.predict_probabilities(input_ids, attention_masks)
        sorted_index = fliplr(argsort(predictions, axis=1))
        return sorted_index[0], predictions[:, sorted_index][0][0]

    def predict_probabilities(self, input_ids, attention_masks):
        self.model.to(self.get_device())
        self.model.eval()
        probabilities = self.model(
            torch.tensor(input_ids, device=self.get_device()),
            attention_mask=torch.tensor(attention_masks, device=self.get_device()),
        )[0]
        probabilities = nn.functional.softmax(probabilities, dim=-1)
        probabilities = probabilities.to(torch.device("cpu"))
        probabilities = probabilities.detach().numpy()
        return probabilities

    def parse(self, message: RuthData):
        input_ids = [message.get(INPUT_IDS)]
        attention_masks = [message.get(ATTENTION_MASKS)]
        index, probabilities = self._predict(input_ids, attention_masks)

        intents = self._change_int_to_text(index)
        probabilities = probabilities

        if intents.size > 0 and probabilities.size > 0:
            ranking = list(zip(list(intents), list(probabilities)))[
                :LABEL_RANKING_LIMIT
            ]
            intent = {"name": intents[0], "accuracy": probabilities[0]}
            intent_rankings = [
                {"name": name, "accuracy": probability} for name, probability in ranking
            ]
        else:
            intent = {"name": None, "accuracy": 0.0}
            intent_rankings = []
        message.set(INTENT, intent)
        message.set(INTENT_RANKING, intent_rankings)


class HFDatasetLoader(Dataset):
    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __len__(self):
        return len(self.labels)

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["labels"] = torch.tensor(self.labels[idx])
        return item



from typing import Any, Dict, List, Text

import numpy as np
from numpy import ndarray
from ruth.nlu.elements import Element
from sklearn.preprocessing import LabelEncoder


class IntentClassifier(Element):
    MAX_LENGTH = 30000

    def __init__(self, element_config: Dict[Text, Any], le: LabelEncoder):
        self.le = le or LabelEncoder()
        super(IntentClassifier, self).__init__(element_config)

    def encode_the_str_to_int(self, labels: List[Text]) -> ndarray:
        return self.le.fit_transform(labels)

    def _change_int_to_text(self, prediction: ndarray) -> ndarray:
        return self.le.inverse_transform(prediction)

    @staticmethod
    def ravel_vector(vector: ndarray) -> ndarray:
        return np.ravel(vector)

    @staticmethod
    def pad_vector(vector: ndarray, max_length: int) -> ndarray:
        if len(vector) < max_length:
            vector = np.pad(vector, (0, max_length - len(vector)), "constant")
        if len(vector) > max_length:
            vector = vector[:max_length]
        return vector

    def get_max_length(self, vector: List[ndarray]):
        max_length = 0
        for i in vector:
            if max_length < i.shape[0]:
                max_length = i.shape[0]
            if max_length > self.MAX_LENGTH:
                return self.MAX_LENGTH
        return max_length

    @staticmethod
    def check_dense(X) -> bool:
        if isinstance(X, ndarray):
            return True
        return False



import io
import os
from typing import Any, Dict, List, Optional, Text
from urllib import request

import numpy
from progressbar import progressbar
from ruth.nlu.classifiers.constants import MODEL_NAME
from ruth.nlu.constants import ELEMENT_UNIQUE_NAME
from ruth.nlu.featurizers.dense_featurizers.dense_featurizer import DenseFeaturizer
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.feature import Feature
from ruth.shared.nlu.training_data.ruth_data import RuthData
from tqdm import tqdm

pbar = None


class FastTextFeaturizer(DenseFeaturizer):
    DO_LOWER_CASE = "do_lower_case"

    defaults = {MODEL_NAME: "wiki-news-300d-1M.vec.zip", DO_LOWER_CASE: True}
    DEFAULT_MODELS_DIR = os.path.join(
        os.path.expanduser("~"), ".cache", "ruth", "models"
    )

    MODELS = {
        "wiki-news-300d-1M.vec.zip": "https://dl.fbaipublicfiles.com/"
        "fasttext/vectors-english/wiki-news-300d-1M.vec.zip",
        "wiki-news-300d-1M-subword.vec.zip": "https://dl.fbaipublicfiles.com/"
        "fasttext/vectors-english/wiki-news-300d-1M-subword.vec.zip",
        "crawl-300d-2M.vec.zip": "https://dl.fbaipublicfiles.com/"
        "fasttext/vectors-english/crawl-300d-2M.vec.zip",
        "crawl-300d-2M-subword.zip": "https://dl.fbaipublicfiles.com/"
        "fasttext/vectors-english/crawl-300d-2M-subword.zip",
    }

    def __init__(self, element_config: Optional[Dict[Text, Any]]):
        super(FastTextFeaturizer, self).__init__(element_config)
        self.vectors = None
        self.featurizer = {}
        if self.element_config[MODEL_NAME] not in self.MODELS:
            raise ValueError(
                "Model name not found. Please choose from the following: "
                "{}".format(list(self.MODELS.keys()))
            )
        self.file_path = self.download_models(self.element_config[MODEL_NAME])
        self.dimension = 300

    def download_models(self, specific_models=None):
        os.makedirs(self.DEFAULT_MODELS_DIR, exist_ok=True)

        def show_progress(block_num, block_size, total_size):
            global pbar
            if pbar is None:
                pbar = progressbar.ProgressBar(maxval=total_size)
                pbar.start()

            downloaded = block_num * block_size
            if downloaded < total_size:
                pbar.update(downloaded)
            else:
                pbar.finish()
                pbar = None

        for model_name, url in self.MODELS.items():
            if specific_models is not None and str(model_name) not in str(
                specific_models
            ):
                continue
            model_path = os.path.join(self.DEFAULT_MODELS_DIR, model_name)
            if os.path.exists(model_path):
                model_path = model_path[:-4]
                return model_path

            request.urlretrieve(url, model_path, show_progress)

            import zipfile

            with zipfile.ZipFile(model_path, "r") as zip_ref:
                zip_ref.extractall(self.DEFAULT_MODELS_DIR)

            model_path = model_path[:-4]
            return model_path
        raise f"""Given model {specific_models} not found.
                Please check the documentation and give the
                right Fastext model name """

    def train(self, training_data: TrainData):
        self.featurizer = self._build_featurizer()
        tokenized_data: List[List[Text]] = [
            message.get_tokenized_data() for message in training_data.intent_examples
        ]

        self.vectors = [
            self.get_vector_list(token_list) for token_list in tokenized_data
        ]

        for message, vector in zip(training_data.training_examples, self.vectors):
            message.add_features(
                Feature(vector, self.element_config[ELEMENT_UNIQUE_NAME])
            )

    def _build_featurizer(self):

        fasttext_corpus = io.open(
            self.file_path, "r", encoding="utf-8", newline="\n", errors="ignore"
        )
        model = {}
        for line in tqdm(fasttext_corpus, colour="red"):
            tokens = line.strip().split(" ")
            model[tokens[0]] = numpy.array(list(tokens[1:]))

        return model

    def get_vector_list(self, token_list):
        if self.featurizer == {}:
            self._build_featurizer()
        if not token_list:
            return numpy.zeros(self.dimension)
        return numpy.array([self.get_vector(token) for token in token_list])

    def get_vector(self, token):
        if token in self.featurizer and self.featurizer != {}:
            return self.featurizer[token]
        else:
            return numpy.zeros(self.dimension)

    def parse(self, message: RuthData):
        tokens = message.get_tokenized_data()
        vector = self.get_vector_list(tokens)
        message.add_features(Feature(vector, self.element_config[ELEMENT_UNIQUE_NAME]))



import copy
from typing import Any, Dict, Optional, Text


def override_defaults(
    defaults: Optional[Dict[Text, Any]], custom: Optional[Dict[Text, Any]]
) -> Dict[Text, Any]:
    if defaults:
        config = copy.deepcopy(defaults)
    else:
        config = {}

    if custom:
        for key in custom.keys():
            if isinstance(config.get(key), dict):
                config[key].update(custom[key])
            else:
                config[key] = custom[key]

    return config



import copy
from typing import Any, Dict, List, Optional, Text

from ruth.constants import INTENT, TEXT, TOKENS
from ruth.shared.nlu.training_data.feature import Feature


class RuthData:
    def __init__(
        self,
        data: Dict[Text, Any] = None,
        features: Optional[List[Feature]] = None,
    ):
        self.features = features or []
        self.data = data or {}

    @classmethod
    def build(cls, intent: Text = None, text: Text = None) -> "RuthData":
        return cls(data={INTENT: intent, TEXT: text})

    def add_features(self, feature: Feature) -> None:
        if feature:
            self.features.append(feature)

    def set(self, key: Text, value: Any):
        self.data[key] = value

    def get(self, key: Text, default: Any = None):
        return self.data.get(key, default)

    @staticmethod
    def _combine_features(features: List[Feature], featurizers: List[Text]) -> Feature:

        combined_features = None

        for feature in features:
            if not combined_features:
                combined_features = copy.deepcopy(feature)
                combined_features.origin = featurizers
            else:
                combined_features.combine_with_features(feature)
        return combined_features

    def get_features(self, featurizers: List[Text] = None) -> Feature:
        if not featurizers and self.features:
            featurizers = [self.features[0].origin]
        combined_features = self._combine_features(self.features, featurizers)
        return combined_features

    def as_dict(self) -> Dict:
        return {key: value for key, value in self.data.items() if value is not None}

    def get_tokenized_data(self) -> List[Text]:
        if self.data.get(TOKENS):
            tokens = [token.text for token in self.data[TOKENS]]
            return tokens
        else:
            return []



DEFAULT_OBJECT_DIRECTORY = "models"
KEY_NAME = "name"
ELEMENT_INDEX = "index"
TOKENS = "tokens"
INTENT_NAME_KEY = "name"
PREDICTED_CONFIDENCE_KEY = "confidence"
ENTITIES = "entities"
INPUT_IDS = "input_ids"
ATTENTION_MASKS = "attention_masks"
# PIPELINE

KEY_PIPELINE = "pipeline"
KEY_LANGUAGE = "language"

DEFAULT_ENCODING = "utf-8"



from typing import Dict, Text

from ruth.nlu.classifiers.hf_classifier import HFClassifier
from ruth.nlu.classifiers.naive_bayes_classifier import NaiveBayesClassifier
from ruth.nlu.classifiers.svm_classifier import SVMClassifier
from ruth.nlu.elements import Element
from ruth.nlu.featurizers.dense_featurizers.fast_text import FastTextFeaturizer
from ruth.nlu.featurizers.sparse_featurizers.count_vector_featurizer import (
    CountVectorFeaturizer,
)
from ruth.nlu.featurizers.sparse_featurizers.tfidf_vector_featurizer import (
    TfidfVectorFeaturizer,
)
from ruth.nlu.tokenizer.hf_tokenizer import HFTokenizer
from ruth.nlu.tokenizer.whitespace_tokenizer import WhiteSpaceTokenizer

element_classes = [
    # Tokenizer
    WhiteSpaceTokenizer,
    HFTokenizer,
    # Featurizers
    CountVectorFeaturizer,
    FastTextFeaturizer,
    TfidfVectorFeaturizer,
    # Classifiers
    HFClassifier,
    NaiveBayesClassifier,
    SVMClassifier,
]

registered_classes: Dict[Text, Element] = {cls.name: cls for cls in element_classes}



from ruth.nlu.constants import ELEMENT_UNIQUE_NAME
from ruth.nlu.featurizers.featurizer import Featurizer


class DenseFeaturizer(Featurizer):
    def __init__(self, element_config):
        element_config = element_config or {}
        self.element_config = element_config
        element_config.setdefault(ELEMENT_UNIQUE_NAME, self.create_unique_name())
        super().__init__(element_config)

    @staticmethod
    def get_default_config():
        pass



from typing import Any, Dict, Text


class Tokens:
    def __init__(self, text: Text, start: int, end: int, data: Dict[Text, Any] = None):
        self.text = text
        self.start = start
        self.end = end
        self.data = data or {}

    def __eq__(self, other):
        return (self.text, self.start, self.end) == (other.text, other.start, other.end)



from typing import Any, Dict, Optional, Text


class RuthModelConfig:
    def __init__(self, configuration_values: Optional[Dict[Text, Any]] = None) -> None:

        if not configuration_values:
            configuration_values = {}

        self.language = "en"
        self.pipeline = []
        self.data = None

        self.override(configuration_values)

        if self.__dict__["pipeline"] is None:
            # replaces None with empty list
            self.__dict__["pipeline"] = []

        for key, value in self.items():
            setattr(self, key, value)



import json
from pathlib import Path
from typing import Any, Text, Union

import yaml
from ruth.shared.constants import DEFAULT_ENCODING


def write_text_file(
    content: Text,
    file_path: Union[Text, Path],
    encoding: Text = DEFAULT_ENCODING,
    append: bool = False,
) -> None:
    mode = "a" if append else "w"
    with open(file_path, mode, encoding=encoding) as file:
        file.write(content)


def json_pickle(file_name: Union[Text, Path], obj: Any, indent: int = 2) -> None:
    import jsonpickle.ext.numpy as jsonpickle_numpy
    import jsonpickle

    jsonpickle_numpy.register_handlers()

    write_text_file(jsonpickle.dumps(obj, indent=indent), file_name)


def read_data(path: Path):
    with open(path, "r") as f:
        return yaml.safe_load(f)


def read_training_data(path: Path):
    with open(path, "r") as f:
        return json.load(f)


def read_file(file_name: Union[Text, Path]):
    try:
        with open(file_name) as f:
            return f.read()
    except FileNotFoundError:
        raise Exception(
            f"Failed to read file, " f"'{str(file_name.absolute())}' does not exist."
        )
    except UnicodeDecodeError:
        raise Exception(
            f"Failed to read file '{file_name.absolute()}', "
            f"Please make sure the file is stored with this "
            f"encoding."
        )


def json_unpickle(file_name: Union[Text, Path]) -> Any:
    import jsonpickle.ext.numpy as jsonpickle_numpy
    import jsonpickle

    jsonpickle_numpy.register_handlers()

    file_content = read_file(file_name)
    return jsonpickle.loads(file_content)



from pathlib import Path
from typing import Any, Dict, List, Optional, Text

from ruth.constants import TEXT
from ruth.nlu.classifiers.constants import MODEL_NAME
from ruth.nlu.tokenizer.constants import MAX_LENGTH_FOR_PADDING
from ruth.nlu.tokenizer.tokenizer import Tokenizer
from ruth.shared.constants import ATTENTION_MASKS, INPUT_IDS
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_data import RuthData
from tqdm import tqdm
from transformers import AutoTokenizer


class HFTokenizer(Tokenizer):
    DO_LOWER_CASE = "do_lower_case"

    defaults = {MODEL_NAME: "bert-base-uncased", DO_LOWER_CASE: True}

    def __init__(self, element_config: Optional[Dict[Text, Any]], tokenizer=None):
        super(HFTokenizer, self).__init__(element_config)
        self.tokenizer = tokenizer or {}

    def _build_tokenizer(self):
        return AutoTokenizer.from_pretrained(self.element_config[MODEL_NAME])

    def _create_tokens(self, examples: TrainData):
        before_padding_text = [message.get(TEXT) for message in examples]

        encoded = self.tokenizer(
            before_padding_text,
            add_special_tokens=True,
            max_length=MAX_LENGTH_FOR_PADDING,
            padding=True,
            truncation=True,
            return_attention_mask=True,
        )

        input_ids = encoded["input_ids"]
        attention_masks = encoded["attention_mask"]
        return input_ids, attention_masks

    def tokenize(self, training_data: TrainData):
        return self._create_tokens(training_data.training_examples)

    @staticmethod
    def _add_tokens_to_data(
        training_examples: List[RuthData],
        input_ids: List[List[int]],
        attention_masks: List[List[int]],
    ):
        for message, input_id, attention_mask in tqdm(
            zip(training_examples, input_ids, attention_masks),
            desc="tokenization",
            total=len(training_examples),
        ):
            message.set(INPUT_IDS, input_id)
            message.set(ATTENTION_MASKS, attention_mask)

    def train(self, training_data: TrainData):
        self.tokenizer = self._build_tokenizer()
        input_ids, attention_masks = self.tokenize(training_data)
        self._add_tokens_to_data(
            training_data.training_examples, input_ids, attention_masks
        )

    def persist(self, file_name: Text, model_dir: Path):
        tokenizer_file_name = file_name + "_tokenizer"

        tokenizer_path = str(model_dir) + "/" + tokenizer_file_name

        if self.tokenizer:
            self.tokenizer.save_pretrained(tokenizer_path)

        return {"tokenizer": tokenizer_file_name}

    @classmethod
    def load(cls, meta: Dict[Text, Any], model_dir: Path, **kwargs):
        tokenizer_file_name = model_dir / meta["tokenizer"]

        tokenizer = AutoTokenizer.from_pretrained(tokenizer_file_name)

        return cls(meta, tokenizer=tokenizer)

    def parse(self, message: RuthData):
        parser_token = self.tokenizer.encode_plus(
            message.get(TEXT),
            add_special_tokens=True,
            max_length=MAX_LENGTH_FOR_PADDING,
            padding=True,
            truncation=True,
            return_attention_mask=True,
        )
        message.set(INPUT_IDS, parser_token["input_ids"])
        message.set(ATTENTION_MASKS, parser_token["attention_mask"])



import copy
import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Text

import ruth
from rich.console import Console
from ruth.constants import INTENT, TEXT
from ruth.nlu.constants import RUTH
from ruth.nlu.elements import Element
from ruth.nlu.registry import registered_classes
from ruth.nlu.utils import check_required_elements, module_path_from_object
from ruth.shared.constants import INTENT_NAME_KEY, PREDICTED_CONFIDENCE_KEY
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_config import RuthConfig
from ruth.shared.nlu.training_data.ruth_data import RuthData
from ruth.shared.utils import json_pickle

logger = logging.getLogger(__name__)
console = Console()


class ElementBuilder:
    def __init__(self, use_cache: bool = True):
        self.use_cache = use_cache

        self.element_cache = {}

    @staticmethod
    def create_element(name: Text, element_config: Dict[Text, Any]):
        if name not in registered_classes:
            logger.error(
                f"Given {name} element is not an registered element. We won't support custom element as of now."
            )
        else:

            return registered_classes[name].build(element_config)

    @staticmethod
    def load_element(name: Text, element_config: Dict[Text, Any], model_dir: Path):
        if name not in registered_classes:
            logger.error(
                f"Given {name} element is not an registered element. We won't support custom element now."
            )
        else:
            return registered_classes[name].load(element_config, model_dir=model_dir)


class MetaData:
    def __init__(self, metadata: Dict[Text, Any]):
        self.metadata = metadata

    def get(self, prop: Text, defaults: Any = None):
        return self.metadata.get(prop, defaults)

    def persist(self, model_dir: Path):
        metadata = self.metadata.copy()

        metadata.update(
            {"Training completed at": datetime.now(), "ruth_version": ruth.VERSION}
        )
        filename = model_dir / "metadata.json"
        json_pickle(filename, metadata, indent=4)


class Trainer:
    def __init__(self, config: RuthConfig, element_builder: ElementBuilder = None):
        self.config = config
        self.training_data = None

        if element_builder is None:
            element_builder = ElementBuilder()

        self.pipeline = self._build_pipeline(config, element_builder)
        self.validate_pipeline()

    def validate_pipeline(self) -> None:
        missing_element = []
        for i, element in enumerate(self.pipeline):
            for required_element in element.required_element():
                if not check_required_elements(required_element, self.pipeline[:i]):
                    missing_element.append(required_element)
                    raise ValueError(
                        f'Missing required elements {" ,".join(missing_element)}'
                    )

    @staticmethod
    def _build_pipeline(config: RuthConfig, element_builder: ElementBuilder):
        pipeline = []

        for index, pipe_element in enumerate(config.pipeline):
            element_cnf = config.get_element(index)
            element = element_builder.create_element(pipe_element["name"], element_cnf)
            pipeline.append(element)

        return pipeline

    @staticmethod
    def get_filename(index, name):
        return f"element_{index}_{name}"

    def train(self, data: TrainData):
        self.training_data = data

        working_data: TrainData = copy.deepcopy(data)

        for i, element in enumerate(self.pipeline):
            console.print(f"Starting to train element {element.name}")
            # component.prepare_partial_processing(self.pipeline[:i], context)
            element.train(working_data)
            console.print(f"Finished training element {element.name}.")
        # return Interpreter(self.pipeline, context)

    def persist(self, path: Path) -> Path:
        timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")

        model_name = RUTH + "_" + timestamp
        metadata = {"language": self.config.language, "pipeline": []}

        model_dir = path.absolute() / model_name
        model_dir.mkdir(exist_ok=True)

        for index, element in enumerate(self.pipeline):
            file_name = self.get_filename(index, name=element.name)
            custom_meta = element.persist(file_name, model_dir)
            element_meta = element.element_config
            if element_meta and custom_meta:
                element_meta.update(custom_meta)
            element_meta["class"] = module_path_from_object(element)

            metadata["pipeline"].append(element_meta)

        MetaData(metadata).persist(model_dir)
        return model_dir


class Interpreter:
    def __init__(
        self,
        pipeline: List[Element],
        model_metadata: Optional[MetaData] = None,
    ) -> None:
        self.pipeline = pipeline
        self.model_metadata = model_metadata

    def parse(
        self,
        text: Text,
    ) -> Dict[Text, Any]:
        """Parse the input text, classify it and return pipeline result.

        The pipeline result usually contains intent and entities."""

        if not text:
            output = self.default_output_attributes()
            output["text"] = ""
            return output

        data = self.default_output_attributes()
        data[TEXT] = text

        message = RuthData(data=data)

        for element in self.pipeline:
            element.parse(message)

        output = self.default_output_attributes()
        output.update(message.as_dict())
        return output

    @staticmethod
    def default_output_attributes():
        return {
            TEXT: "",
            INTENT: {INTENT_NAME_KEY: None, PREDICTED_CONFIDENCE_KEY: 0.0},
        }



import copy
from typing import Any, Dict, List, Text

from ruth.nlu.constants import INDEX
from ruth.shared.constants import ELEMENT_INDEX, KEY_LANGUAGE


@property
def default_pipline() -> List[Dict[Text, Any]]:
    return [{"name": "CountVectorFeaturizer"}, {"name": "NaiveBayesClassifier"}]


class RuthConfig:
    def __init__(self, config: Dict[Text, Any]):
        self.config = config or {}

        self.language = config.get(KEY_LANGUAGE, "en")
        self.pipeline = []

        for key, value in config.items():
            setattr(self, key, value)

    @staticmethod
    def index_the_pipeline(pipeline: List[Dict[Text, Any]]):
        if not pipeline:
            return []

        return [
            element.setdefault(ELEMENT_INDEX, index)
            for index, element in enumerate(pipeline)
        ]

    def get_element(
        self, index: int, default: Dict[Text, Any] = None
    ) -> Dict[Text, Any]:
        try:
            component = copy.deepcopy(self.pipeline[index])
            component[INDEX] = index
            return component
        except IndexError:
            if default:
                default[INDEX] = index
            else:
                default = {INDEX: index}
            return default

    def __getitem__(self, key: Text) -> Any:
        return self.config[key]



from typing import Any, Dict, List, Text

from ruth.nlu.featurizers.featurizer import Featurizer
from ruth.nlu.featurizers.sparse_featurizers.constants import (
    CLASS_FEATURIZER_UNIQUE_NAME,
)
from ruth.shared.nlu.training_data.collections import TrainData


class SparseFeaturizer(Featurizer):
    def __init__(self, element_config):
        element_config = element_config or {}
        self.element_config = element_config
        element_config.setdefault(
            CLASS_FEATURIZER_UNIQUE_NAME, self.create_unique_name()
        )
        super().__init__(element_config)

    def _build_vectorizer(self, parameters: Dict[Text, Any]):
        raise NotImplementedError

    @staticmethod
    def get_data(training_data: TrainData) -> List[Text]:
        return training_data.get_text_list(
            training_examples=training_data.training_examples
        )



import json
from pathlib import Path
from typing import List, Text

from ruth.constants import INTENT, TEXT
from ruth.shared.nlu.training_data.ruth_data import RuthData


class TrainData:
    def __init__(self, training_examples: List[RuthData] = None):
        self.training_examples = training_examples or []

    def __len__(self) -> int:
        return len(self.training_examples)

    @classmethod
    def build(cls, data_path: Path) -> "TrainData":
        training_examples = list()
        with open(data_path, "r") as f:
            messages = json.load(f)
        for message in messages:
            training_examples.append(
                RuthData.build(intent=message.get(INTENT), text=message.get(TEXT))
            )
        return cls(training_examples)

    def add_example(self, data: RuthData) -> List[RuthData]:
        self.training_examples.append(data)
        return self.training_examples

    def __getitem__(self, item):
        return self.training_examples[item]

    @staticmethod
    def get_text_list(training_examples: List[RuthData]) -> List[Text]:
        return [example.get(TEXT) for example in training_examples]

    @property
    def intent_examples(self) -> List[RuthData]:
        """Returns the list of examples that have intent."""
        return [ex for ex in self.training_examples if ex.get(INTENT)]



import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Text

from rich.console import Console
from ruth.constants import TEXT
from ruth.nlu.featurizers.sparse_featurizers.constants import (
    CLASS_FEATURIZER_UNIQUE_NAME,
)
from ruth.nlu.featurizers.sparse_featurizers.sparse_featurizer import SparseFeaturizer
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.feature import Feature
from ruth.shared.nlu.training_data.ruth_data import RuthData
from ruth.shared.utils import json_pickle, json_unpickle
from sklearn.feature_extraction.text import CountVectorizer

logger = logging.getLogger(__name__)

console = Console()


class CountVectorFeaturizer(SparseFeaturizer):
    defaults = {
        "analyzer": "word",
        "stop_words": None,
        "min_df": 1,
        "max_df": 1.0,
        "min_ngram": 1,
        "max_ngram": 1,
        "lowercase": True,
        "max_features": None,
        "use_lemma": True,
    }

    def _load_params(self):
        self.analyzer = self.element_config["analyzer"]
        self.stop_words = self.element_config["stop_words"]
        self.min_df = self.element_config["min_df"]
        self.max_df = self.element_config["max_df"]
        self.min_ngram = self.element_config["min_ngram"]
        self.max_ngram = self.element_config["max_ngram"]
        self.lowercase = self.element_config["lowercase"]
        self.use_lemma = self.element_config["use_lemma"]

    def _verify_analyzer(self) -> None:
        if self.analyzer != "word":
            if self.stop_words is not None:
                logger.warning(
                    "You specified the character wise analyzer."
                    " So stop words will be ignored."
                )
            if self.max_ngram == 1:
                logger.warning(
                    "You specified the character wise analyzer"
                    " but max n-gram is set to 1."
                    " So, the vocabulary will only contain"
                    " the single characters. "
                )

    def __init__(
        self,
        element_config: Optional[Dict[Text, Any]],
        vectorizer: Optional["CountVectorizer"] = None,
    ):
        super(CountVectorFeaturizer, self).__init__(element_config)
        self.vectorizer = vectorizer
        self._load_params()
        self._verify_analyzer()

    @staticmethod
    def _build_vectorizer(
        parameters: Dict[Text, Any], vacabulary=None
    ) -> CountVectorizer:
        return CountVectorizer(
            analyzer=parameters["analyzer"],
            stop_words=parameters["stop_words"],
            min_df=parameters["min_df"],
            max_df=parameters["max_df"],
            ngram_range=(parameters["min_ngram"], parameters["max_ngram"]),
            lowercase=parameters["lowercase"],
            vocabulary=vacabulary,
        )

    def _check_attribute_vocabulary(self) -> bool:
        """Checks if trained vocabulary exists in attribute's count vectorizer."""
        try:
            return hasattr(self.vectorizer, "vocabulary_")
        except (AttributeError, KeyError):
            return False

    def create_vectors(self, examples: List[RuthData]):
        features = []
        for message in examples:
            features.append(self.vectorizer.transform([message.get(TEXT)]))
        return features

    def _get_featurizer_data(self, training_data: TrainData):
        if self._check_attribute_vocabulary():
            return self.create_vectors(training_data.training_examples)
        else:
            return []

    def _add_features_to_data(self, training_examples: List[RuthData], features):
        for message, feature in zip(training_examples, features):
            message.add_features(
                Feature(feature, self.element_config[CLASS_FEATURIZER_UNIQUE_NAME])
            )

    def train(self, training_data: TrainData) -> CountVectorizer:
        self.vectorizer = self._build_vectorizer(
            parameters={
                "analyzer": self.analyzer,
                "stop_words": self.stop_words,
                "min_df": self.min_df,
                "max_df": self.max_df,
                "min_ngram": self.min_ngram,
                "max_ngram": self.max_ngram,
                "lowercase": self.lowercase,
            }
        )
        self.vectorizer.fit(self.get_data(training_data))
        features = self._get_featurizer_data(training_data)
        self._add_features_to_data(training_data.training_examples, features)
        return self.vectorizer

    def parse(self, message: RuthData):
        feature = self.vectorizer.transform([message.get(TEXT)])
        message.add_features(
            Feature(feature, self.element_config[CLASS_FEATURIZER_UNIQUE_NAME])
        )

    def get_vocablary_from_vectorizer(self):
        if self.vectorizer.vocabulary_:
            return self.vectorizer.vocabulary_
        else:
            raise "CountVectorizer not got trained. Please check the training data and retrain the model"

    def persist(self, file_name: Text, model_dir: Text):
        file_name = file_name + ".pkl"
        if self.vectorizer:
            vocab = self.vectorizer.vocabulary_

            featurizer_path = Path(model_dir) / file_name
            json_pickle(featurizer_path, vocab)

        return {"file_name": file_name}

    @classmethod
    def load(
        cls, meta: Dict[Text, Any], model_dir: Path, **kwargs: Any
    ) -> "CountVectorFeaturizer":
        file_name = meta.get("file_name")
        featurizer_file = model_dir / file_name

        if not featurizer_file.exists():
            return cls(meta)

        vocabulary = json_unpickle(featurizer_file)

        vectorizers = cls._build_vectorizer(parameters=meta, vacabulary=vocabulary)

        return cls(meta, vectorizers)



from typing import Any, Dict, List, Text

from ruth.constants import TEXT, TOKENS
from ruth.nlu.constants import ELEMENT_UNIQUE_NAME
from ruth.nlu.elements import Element
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_data import RuthData


class Token:
    def __init__(self, text: Text, start: int, end: int, data: Dict[Text, Any] = None):
        self.text = text
        self.start = start
        self.end = end
        self.data = data or {}

    def __eq__(self, other):
        return (self.text, self.start, self.end) == (other.text, other.start, other.end)


class Tokenizer(Element):
    def __init__(self, element_config):
        element_config = element_config or {}
        self.element_config = element_config
        element_config.setdefault(ELEMENT_UNIQUE_NAME, self.create_unique_name())
        super().__init__(element_config)

    def train(self, training_data: TrainData):
        for data in training_data.training_examples:
            text = data.get(TEXT)
            tokens = self.tokenize(text)
            data.set(TOKENS, tokens)

    def parse(self, message: RuthData):
        tokens = self.tokenize(message.get(TEXT))
        message.set(TOKENS, tokens)

    @staticmethod
    def _convert_words_to_tokens(words: List[Text], text: Text) -> List[Token]:
        running_offset = 0
        tokens = []

        for word in words:
            word_offset = text.index(word, running_offset)
            word_len = len(word)
            running_offset = word_offset + word_len
            tokens.append(Token(word.lower(), start=word_offset, end=running_offset))

        return tokens

    def tokenize(self, text: Text):
        raise NotImplementedError(
            f"failed to implement the tokenize function in {self.name}"
        )



#!/usr/bin/env python3
import pathlib

import setuptools
from ruth import VERSION


here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

core_requirements = [
    "click~=8.0.1",
    "rich~=12.0.0",
    "aiohttp~=3.6.3",
    "jsonpickle~=2.1.0",
    "pyyaml~=6.0",
    "torch~=1.12.1",
    "regex",
    "scikit-learn~=1.0.2",
    "matplotlib~=3.5.3",
    "progressbar~=2.5",
    "transformers~=4.20.0",
]


setuptools.setup(
    name='ruth-py',
    description="A Python CLI for Ruth NLP",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="",
    author='Puretalk',
    author_email='info@puretalk.ai',
    version=VERSION,
    install_requires=core_requirements,
    python_requires='>=3.8,<=3.10',
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    include_package_data=True,
    package_data={
        "data": ["*.txt"]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3.8',
    ],
    entry_points={"console_scripts": ["ruth = ruth.cli.cli:entrypoint"]},
)


import json
import os
import re
from pathlib import Path
from typing import Text

import click
import matplotlib.pyplot as plt
from rich.console import Console
from ruth import VERSION
from ruth.cli.utills import (
    build_pipeline_from_metadata,
    get_config,
    get_metadata_from_model,
)
from ruth.constants import INTENT
from ruth.nlu.model import Interpreter
from ruth.nlu.train import train_pipeline
from ruth.shared.nlu.training_data.collections import TrainData
from ruth.shared.nlu.training_data.ruth_config import RuthConfig
from sklearn.metrics import confusion_matrix

console = Console()


class RichGroup(click.Group):
    def format_help(self, ctx, formatter):
        # TODO: Want to write the help description whenever the user call the ruth --help
        ...


@click.group(cls=RichGroup)
@click.version_option(VERSION)
def entrypoint():
    pass


@entrypoint.command(name="train")
@click.option(
    "-d",
    "--data",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Data for training as json",
)
@click.option(
    "-p",
    "--pipeline",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="pipeline for training as yaml",
)
def train(data: Path, pipeline: Path):
    config = get_config(pipeline)
    training_data = TrainData.build(data)

    config = RuthConfig(config)
    model_absolute_dir = train_pipeline(config, training_data)
    console.print(
        f"Training is completed and model is stored at [yellow]{model_absolute_dir}[/yellow]"
    )


@entrypoint.command(name="parse")
@click.option(
    "-t",
    "--text",
    type=click.STRING,
    required=True,
    help="Data that need to be get parsed",
)
@click.option(
    "-m",
    "--model_path",
    type=click.STRING,
    default="models",
    help="Directory where the model is stored",
)
def parse(text: Text, model_path: Text):
    models = [
        directory
        for directory in Path(model_path).iterdir()
        if directory.is_dir() and re.search("ruth", str(directory))
    ]
    models.sort()

    latest_model = models[-1]

    console.print(f"Latest Model found {latest_model}")
    metadata = get_metadata_from_model(latest_model.absolute())
    pipeline = build_pipeline_from_metadata(metadata=metadata, model_dir=latest_model)
    interpreter = Interpreter(pipeline)
    output = interpreter.parse(text)
    console.print(f"Predicted intent is {output.get(INTENT)}")


@entrypoint.command(name="evaluate")
@click.option(
    "-d",
    "--data",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Data for testing as json",
)
@click.option(
    "-m",
    "--model_path",
    type=click.Path(exists=True),
    default=Path("models"),
    help="Directory where the model is stored",
)
@click.option(
    "-o",
    "--output_folder",
    type=click.Path(),
    default=Path("results"),
    help="Directory where the results is stored",
)
def evaluate(data: Path, model_path: Text, output_folder: Text):
    models = [
        directory
        for directory in Path(model_path).iterdir()
        if directory.is_dir() and re.search("ruth", str(directory))
    ]
    models.sort()

    latest_model = models[-1]

    console.print(f"Latest Model found {latest_model}")
    metadata = get_metadata_from_model(latest_model.absolute())
    pipeline = build_pipeline_from_metadata(metadata=metadata, model_dir=latest_model)
    interpreter = Interpreter(pipeline)
    with open(data, "r") as f:
        examples = json.load(f)

    correct_predictions = 0
    y_pred = []
    y_actual = []
    for example in examples:
        output = interpreter.parse(example["text"])

        y_pred.append(output.get(INTENT).get("name"))
        y_actual.append(example["intent"])

        if output.get(INTENT).get("name") == example["intent"]:
            correct_predictions += 1

    accuracy = correct_predictions / len(examples)
    conf_matrix = confusion_matrix(y_true=y_actual, y_pred=y_pred)

    fig, ax = plt.subplots(figsize=(7.5, 7.5))
    ax.matshow(conf_matrix, cmap=plt.cm.Blues, alpha=0.3)
    for i in range(conf_matrix.shape[0]):
        for j in range(conf_matrix.shape[1]):
            ax.text(
                x=j, y=i, s=conf_matrix[i, j], va="center", ha="center", size="xx-large"
            )

    plt.xlabel("Predictions", fontsize=18)
    plt.ylabel("Actual", fontsize=18)
    plt.title("Confusion Matrix", fontsize=18)

    if output_folder:
        result_path = Path(output_folder).absolute()
    else:
        result_path = Path().absolute() / output_folder
    result_path.mkdir(exist_ok=True)
    directories = os.listdir(str(result_path))
    indexes = []
    model_name = str(latest_model).split("/")[-1]
    for result in directories:
        if model_name in result:
            indexes.append(int(result.split("@")[-1]))
    if indexes:
        index = max(indexes) + 1
    else:
        index = 0

    folder_for_the_result = result_path / f"{model_name}@{index}"
    folder_for_the_result.mkdir(exist_ok=True)
    final_file_path = folder_for_the_result / "confusion_matrix.png"
    plt.savefig(final_file_path)

    print("accuracy: ", accuracy)
    print("confusion matrix is created.")
    print("results are stored here: ", folder_for_the_result)



from typing import List, Text

from ruth.nlu.elements import Element


def module_path_from_object(o: Element) -> Text:
    return o.__class__.__module__ + "." + o.__class__.__name__


def check_required_elements(required_element: Element, pipeline: List[Element]) -> bool:
    for element in pipeline:
        if isinstance(element, required_element):
            return True
    return False



import json
from pathlib import Path
from typing import Any, Dict, List, Text

import yaml
from ruth.nlu.model import ElementBuilder


def get_config(pipeline_path: Path) -> Dict[Text, Any]:
    with open(pipeline_path, "r") as f:
        return yaml.safe_load(f)


def load_json_data(path: Path) -> Dict[Text, Any]:
    with open(path, "r") as f:
        return json.load(f)


def build_pipeline_from_metadata(
    metadata: Dict[Text, Any],
    model_dir: Path,
    element_builder: ElementBuilder = None,
):
    pipeline_element = []
    if not element_builder:
        element_builder = ElementBuilder()

    pipeline: List[Dict[Text, Any]] = metadata["pipeline"]
    for element in pipeline:
        pipeline_element.append(
            element_builder.load_element(element["name"], element, model_dir=model_dir)
        )
    return pipeline_element


def get_metadata_from_model(model_path: Path) -> Dict[Text, Any]:
    metadata_file_path = model_path / "metadata.json"
    metadata = load_json_data(metadata_file_path)
    return metadata
